#include "include/mns.h"
#include "include/inetraw.h"

char *checkosmain(void);
void show_print(struct node **);

/* global varibles */
extern time_t timedelay;
struct fingerprint pak[PACKETS];

/* local structures */
struct fingerprint {
   u_short set:2,
           seq:2,
           ackset:3,
           urg:1;
   u_long ack;
   u_short win;
   u_short flag;
};

struct node *checkos (struct in_addr target,
                      struct node *walk,
                      int *plist,
                      struct in_addr hostaddress)
{
   char *buffer = NULL;
   int i, pcount;
   char ham_asc[2];
   char *e_msg;

   if (plist != NULL) pcount = plist[0] + 1;
   else return walk;

   if((e_msg = init_packet_devices(IPPROTO_TCP))) {
      stralloc(walk, e_msg);
      free(e_msg);
      return walk;
   }
   
   local_ip = hostaddress;

   for (i = 1; i < min(pcount, 3); i++) {
      dst_port = htons(plist[i]);
      buffer = checkosmain();
      if (buffer != NULL) {
         ham_asc[0] = buffer[0];
	 ham_asc[1] = '\0';
	 stralloc (walk, "\n\tTCP Sequence Prediction Difficulty ->");
         stralloc (walk, ham_asc);
	 switch (ham_asc[0]) {
            case '0':
               stralloc (walk, "( indeterminate )");
	       break;
            case '1':
	       stralloc (walk, "( easy )");
	       break;
	    case '2':
	       stralloc (walk, "( easy )");
               break;
	    case '3':
	       stralloc (walk, "( moderate )");
	       break;
	    case '4':
	       stralloc (walk, "( moderate )");
	       break;
	    case '5':
	       stralloc (walk, "( hard )");
	       break;
	    case '6':
	       stralloc (walk, "( infeaseable )");
	       break;
	    case '7':
	       stralloc (walk, "( impossible )");
	       break;
	    case '8':
	       stralloc (walk, "( impossible )");
	       break;
	    default:
	       stralloc (walk, "( hamming error )");
	       break;
	    }
	    stralloc (walk, "\n\tOperating System ->");
	    stralloc (walk, &buffer[1]);
	    break;
         }
         if (buffer != NULL)
	   break;
      }

   if (buffer != NULL) free(buffer);
   else {
      stralloc(walk, "\n\tOperating System -> ( indeterminate )");
      if(mns_opts & VERBOSE) show_print(&walk);
   }

  term_packet_devices();

   return walk;
}

char *checkosmain(void) {
  FILE *fs_cfg;
  char *osname;
  struct tcphdr tcp;
  struct ip ip;
  hdrinfo s_ack;
  u_short start, s;
  u_long myseq[PACKETS];
  u_long seq_mask = 0, last_seq = 0;
  time_t timeout;
  char line[60];
  int n, i, ham;

  for (n = 0; n < PACKETS; n++)
    bzero (&pak[n], sizeof(struct fingerprint));

  srand(time (NULL) & 0x0000ffff);
  start = s = (rand() % 5000) + 5000;
  s_ack.ack = 0;

#define sendtcppacket(a, b)\
   {s_ack.seq = myseq[a] = rand();\
    src_port = htons(s + a);\
    sendtcp(&s_ack, b);\
    usleep(10);}

  sendtcppacket (0, TH_SYN);
  sendtcppacket (1, TH_SYN | TH_ACK);
  sendtcppacket (2, TH_FIN | TH_ACK);
  sendtcppacket (3, TH_SYN | TH_FIN);
  sendtcppacket (4, TH_SYN | TH_XXX | TH_YYY);
  sendtcppacket (5, TH_SYN); /* this extra syn is used for hamming distances
                              * an even number of responses are needed */
  timeout = time (NULL) + timedelay;
  n = i = 0;

  do {
     if (get_packet(&ip, (char *)&tcp, sizeof(tcp), SRC)) {
        if (dst_port == tcp.th_sport) {
           if (ntohs(tcp.th_dport) - start < PACKETS) {
              n = ntohs(tcp.th_dport) - start;
              if ((n < 0) || (n >= PACKETS)) continue;
              if (pak[n].set) continue;
              if (!n && !ntohs(tcp.th_win)) return NULL;
              pak[n].seq = tcp.th_seq != 0;
              if (pak[n].seq) {
                 seq_mask ^= tcp.th_seq;
                 last_seq = tcp.th_seq;
                 i++;
              }
              pak[n].ack = ntohl (tcp.th_ack);
              pak[n].ackset = tcp.th_ack != 0;
              if (pak[n].ackset && (myseq[n] > pak[n].ack))
                 pak[n].ackset = 2;
              if (pak[n].ackset)
                 pak[n].ack = (myseq[n] < pak[n].ack) ?
                 (pak[n].ack - myseq[n]) :
                 (myseq[n] - pak[n].ack);
              if (pak[n].ack > ACK_RAND) {
                 pak[n].ackset = 3;
                 pak[n].ack = 0;
              }
              pak[n].win = ntohs (tcp.th_win);
              pak[n].flag = tcp.th_flags;
	      pak[n].set = 1;
              pak[n].urg = tcp.th_urp != 0;
           }
        }
     }
     usleep (10);
  }
  while ((timeout > time (NULL)) && (n != PACKETS - 1));
  
  /* calculate hamming distance */
  if (i % 2) seq_mask ^= last_seq;
  ham = hamming (seq_mask);
  if (ham > 8) ham = 16 - ham;

  /* parse fingerprints */
  if ((fs_cfg = fopen (OSCFGFILE, "r")) != NULL)
    {
      struct fingerprint parse;
      int flag1 = 0, found = 0, total = 0;
      char *token;

      osname = malloc (60);

      while(fgets (line, 60, fs_cfg)) {
         if (line[0] == '#') continue; /* COMMENT LINE */
         if (line[0] == '\n') { /* NEW LINE */
            if (flag1 && (found == total)) {
               fclose(fs_cfg);
               osname[0] = (char)ham + '0';
               return osname;
            }
         }
         
         if (line[0] == '*') { /* OS FINGERPRINT HEADER */
            if (strlen (line) > 3) {
               strcpy (osname, &line[1]);
               osname[strlen (osname) - 1] = '\0';
            }
            flag1 = 1;
            total = found = 0;
            continue;
         }

         token = strtok (line, " ");
         if (isdigit (token[0]))
            n = atoi (token);
         else {
            flag1 = total = found = 0;
	    continue;
         }

         if ((token = strtok (NULL, " ")) != NULL)
            parse.seq = atoi (token);
         parse.set = isdigit (token[0]) ? 1 : 0;

	  if ((token = strtok (NULL, " ")) != NULL)
	    {
	      switch (token[0])
		{
		case '+':
		  parse.ackset = 1;
		  parse.ack = atol (&token[1]);
		  break;
		case '-':
		  parse.ackset = 2;
		  parse.ack = atol (&token[1]);
		  break;
		case 'R':
		  parse.ackset = 3;
		  parse.ack = 0;
		  break;
		case 'N':
		default:
		  parse.ackset = 0;
		  parse.ack = 0;
		  break;
		}
	    }

	  if ((token = strtok (NULL, " ")) != NULL)
	    parse.win = strtol (token, NULL, 16);

	  if ((token = strtok (NULL, " \n")) != NULL)
	    {
	      parse.flag = 0;
	      if (strchr (token, 'S'))
		parse.flag |= TH_SYN;
	      if (strchr (token, 'R'))
		parse.flag |= TH_RST;
	      if (strchr (token, 'A'))
		parse.flag |= TH_ACK;
	      if (strchr (token, 'F'))
		parse.flag |= TH_FIN;
	      if (strchr (token, 'P'))
		parse.flag |= TH_PUSH;
	      if (strchr (token, 'X'))
		parse.flag |= TH_XXX;
	      if (strchr (token, 'Y'))
		parse.flag |= TH_YYY;
	      if (strchr (token, 'U'))
		parse.urg = 1;
	      else
		parse.urg = 0;
	    }

	  total++;

	  /* test for no response */
	  if (!pak[n].set)
	    {
	      if (parse.set)
		total = found = flag1 = 0;
	      else
		found++;
	      continue;
	    }

	  if ((parse.ackset == pak[n].ackset) &&
	      (parse.ack == pak[n].ack) &&
	      (parse.seq == pak[n].seq) &&
	      (parse.win == pak[n].win) &&
	      (parse.flag == pak[n].flag) &&
	      (parse.urg == pak[n].urg))
	    {
	      found++;
	      continue;
	    }
	  else
	    {
	      total = found = flag1 = 0;
	      continue;
	    }
	}

      fseek(fs_cfg, 0L, SEEK_END);
      fclose(fs_cfg);
      if(osname != NULL) free(osname);
      return NULL;
    }

  puts("error: cannot open " OSCFGFILE);
  exit(EXIT_FAILURE);
}

void show_print(struct node **walk_p) {
   char buffer[40];
   int i;
   
   stralloc(*walk_p, "\n\t\t* Unknown OS Fingerprint");
   for(i = 0; i < PACKETS - 1; i++) {
      if(pak[i].set) {
         sprintf(buffer, "\n\t\t%i %i ", i, pak[i].seq);
         switch(pak[i].ackset) {
            case 0: strcat(buffer, "N"); break;
            case 1:
               strcat(buffer, "+");
               sprintf(&buffer[strlen(buffer)], "%li", pak[i].ack);
               break;
            case 2:
               strcat(buffer, "-");
               sprintf(&buffer[strlen(buffer)], "%li", pak[i].ack);
               break;
            case 3: strcat(buffer, "R"); break;
         }
         sprintf(&buffer[strlen(buffer)], " %04x ", pak[i].win);
         if(pak[i].flag & TH_SYN) strcat(buffer, "S");
         if(pak[i].flag & TH_RST) strcat(buffer, "R");
         if(pak[i].flag & TH_ACK) strcat(buffer, "A");
         if(pak[i].flag & TH_FIN) strcat(buffer, "F");
         if(pak[i].flag & TH_PUSH) strcat(buffer, "P");
         if(pak[i].flag & TH_XXX) strcat(buffer, "X");
         if(pak[i].flag & TH_YYY) strcat(buffer, "Y");
         if(pak[i].urg) strcat(buffer, "U");
      } else sprintf(buffer, "\n\t\t%i - - - -", i);
      stralloc(*walk_p, buffer);
   }
}
